use std::io;
use std::fs;
use std::path::PathBuf;
use crate::utils::path;
use crate::store::head;
use crate::store::object::{add_node, rm_node};
use crypto::sha1::Sha1;
use crypto::digest::Digest;
use crate::utils::into::IntoPathBuf;
use crate::store::object::{blob::Blob, tree::Tree};
use crate::commands::status::{State, LocalObj};

#[derive(Clone, Copy)]
pub enum ObjType {
    TREE,
    BLOB,
    DEFAULT
}

pub trait ObjMethods {
    fn get_type(&self) -> ObjType;
    fn get_obj_path(&self) -> PathBuf;
    fn get_file_path(&self) -> PathBuf;
    fn get_relative_file_path(&self) -> PathBuf;
    fn get_repo_file_path(&self) -> PathBuf;
    fn get_name(&self) -> String;
    fn get_hash_path(&self) -> String;
    fn get_local_obj(&self) -> LocalObj;
    fn get_line(&self) -> String;
    fn add_ref_to_parent(&self) -> io::Result<()>;
    fn rm(&mut self) -> io::Result<()>;
    fn rm_node(&mut self) -> io::Result<()>;
    fn rm_node_down(&mut self) -> io::Result<()>;
    fn exists_on_remote(&mut self) -> bool;
}

pub struct Obj {
    name: String,
    obj_path: PathBuf,
    obj_type: ObjType,
    file_path: PathBuf, // file here is used as both file and directory
    relative_file_path: PathBuf,
    repo_file_path: PathBuf,
    hash_path: String, // hash of the relative path of the file
}


impl ObjMethods for Obj {
    fn get_type(&self) -> ObjType {
        self.obj_type
    }

    fn get_obj_path(&self) -> PathBuf {
        self.obj_path.clone()
    }

    fn get_file_path(&self) -> PathBuf {
        self.file_path.clone()
    }

    fn get_relative_file_path(&self) -> PathBuf {
        self.relative_file_path.clone()
    }

    fn get_repo_file_path(&self) -> PathBuf {
        self.repo_file_path.clone()
    }

    fn get_local_obj(&self) -> LocalObj {
        LocalObj {
            otype: match self.obj_type {
                ObjType::BLOB => String::from("blob"),
                ObjType::TREE => String::from("tree"),
                ObjType::DEFAULT => String::from("default"),
            },
            name: self.get_name(),
            path: self.get_repo_file_path(),
            path_from: None,
            state: State::New
        }
    }

    fn get_name(&self) -> String {
        self.name.clone()
    }

    fn get_hash_path(&self) -> String {
        self.hash_path.clone()
    }

    // build line for parent reference
    fn get_line(&self) -> String {
        format!("tree {} {}", self.get_hash_path(), self.get_name())
    }

    fn add_ref_to_parent(&self) -> io::Result<()> {
        let line  = self.get_line();
        if self.get_relative_file_path().iter().count() == 1 {
            head::add_line(line)?;
        } else {
            add_node(self.get_relative_file_path().parent().unwrap(), &line)?;
        }
        Ok(())
    }

    fn rm_node(&mut self) -> io::Result<()> {
        // remove self object and children object
        self.rm_node_down();

        // remove parent reference to self
        let line = self.get_line();
        if self.get_relative_file_path().iter().count() == 1 {
            head::rm_line(&line)?;
        } else {
            rm_node(self.get_relative_file_path().parent().unwrap(), &line)?;
        }
        
        Ok(())
    }

    fn rm_node_down(&mut self) -> io::Result<()> {
        eprintln!("rm_node_down: tried to do this on Obj");
        Ok(())
    }

    fn rm(&mut self) -> io::Result<()> {
        eprintln!("rm: tried to do this on Obj");
        Ok(())
    }

    fn exists_on_remote(&mut self) -> bool {
        self.obj_path.exists()
    }
}

impl ObjMethods for Blob {
    fn get_type(&self) -> ObjType {
        self.obj.get_type()
    }

    fn get_obj_path(&self) -> PathBuf {
        self.obj.get_obj_path()
    }

    fn get_file_path(&self) -> PathBuf {
        self.obj.get_file_path()
    }

    fn get_relative_file_path(&self) -> PathBuf {
        self.obj.get_relative_file_path()
    }

    fn get_repo_file_path(&self) -> PathBuf {
        self.obj.get_repo_file_path()
    }

    fn get_local_obj(&self) -> LocalObj {
        self.obj.get_local_obj()
    }

    fn get_name(&self) -> String {
        self.obj.get_name()
    }

    fn get_hash_path(&self) -> String {
        self.obj.get_hash_path()
    }

    fn get_line(&self) -> String {
        self.obj.get_line()
    }

    fn add_ref_to_parent(&self) -> io::Result<()> {
        self.obj.add_ref_to_parent()
    }

    fn rm_node(&mut self) -> io::Result<()> {
        self.obj.rm_node()
    }

    fn rm_node_down(&mut self) -> io::Result<()> {
        // remove reference to self
        fs::remove_file(self.get_obj_path())?;
        Ok(())
    }

    fn rm(&mut self) -> io::Result<()> {
        // remove all references, including children's one
        self.rm_node()?;

        // remove file
        fs::remove_file(self.get_file_path())?;
        Ok(())
    }

    fn exists_on_remote(&mut self) -> bool {
        self.obj.exists_on_remote()
    }
}

impl ObjMethods for Tree {
    fn get_type(&self) -> ObjType {
        self.obj.get_type()
    }

    fn get_obj_path(&self) -> PathBuf {
        self.obj.get_obj_path()
    }

    fn get_file_path(&self) -> PathBuf {
        self.obj.get_file_path()
    }

    fn get_relative_file_path(&self) -> PathBuf {
        self.obj.get_relative_file_path()
    }

    fn get_repo_file_path(&self) -> PathBuf {
        self.obj.get_repo_file_path()
    }

    fn get_local_obj(&self) -> LocalObj {
        self.obj.get_local_obj()
    }

    fn get_name(&self) -> String {
        self.obj.get_name()
    }

    fn get_hash_path(&self) -> String {
        self.obj.get_hash_path()
    }

    fn get_line(&self) -> String {
        self.obj.get_line()
    }

    fn add_ref_to_parent(&self) -> io::Result<()> {
        self.obj.add_ref_to_parent()
    }

    fn rm_node(&mut self) -> io::Result<()> {
        self.obj.rm_node()
    }

    /// remove objects and children but not parent reference to self
    fn rm_node_down(&mut self) -> io::Result<()> {
        // remove children
        while let Some(mut child) = self.next() {
            match child.get_type() {
                ObjType::TREE => child.rm_node_down(),
                ObjType::BLOB => child.rm_node_down(),
                _ => Ok(())
            }?;
        };

        // remove reference to self
        fs::remove_file(self.get_obj_path())?;
        Ok(())
    }

    fn rm(&mut self) -> io::Result<()> {
        // remove all references, including children's one
        self.rm_node()?;

        // remove directory and all subfiles
        fs::remove_dir_all(self.get_file_path())?;
        Ok(())
    }

    fn exists_on_remote(&mut self) -> bool {
        self.obj.exists_on_remote()
    }
}

impl Obj {
    fn new() -> Self {
        Obj { 
            name: String::new(),
            obj_path: PathBuf::new(),
            file_path: PathBuf::new(),
            obj_type: ObjType::DEFAULT,
            hash_path: String::new(),
            relative_file_path: PathBuf::new(),
            repo_file_path: PathBuf::new()
        }
    }

    pub fn from_path<S>(path: S) -> Obj where S: IntoPathBuf {

        let path = path.into();
        let mut hasher = Sha1::new();
        hasher.input_str(path.to_str().unwrap());
        let hash = hasher.result_str();

        let (dir, res) = hash.split_at(2);
        let mut obj_path = path::objects();
        obj_path.push(dir);
        obj_path.push(res);

        // set to absolute path if not already
        let root = path::repo_root();
        let abs_path = match path.clone().starts_with(root.clone()) {
            true => path.clone(),
            false => root.join(path.clone())
        };

        Obj { 
            name: match abs_path.file_name() {
                None => String::new(),
                Some(name) => name.to_str().unwrap().to_owned()
            },
            obj_path,
            obj_type: match path.exists() {
                true => match path.is_dir() {
                    true => ObjType::TREE,
                    false => ObjType::BLOB
                },
                false => ObjType::DEFAULT
            },
            file_path: abs_path,
            relative_file_path: path.clone(),
            repo_file_path: path,
            hash_path: hash,
        }
    }

    /// load from the information line stored in the object
    pub fn from_line(line: String, base_dir: Option<PathBuf>) -> Box<dyn ObjMethods> {
        let mut split = line.trim().rsplit(' ');
        if split.clone().count() != 3 {
            eprintln!("fatal: invalid object(s)");
            std::process::exit(1);
        }

        let name = split.next().unwrap();
        let hash_path = split.next().unwrap();
        let obj_type = split.next().unwrap();

        let (dir, res) = hash_path.split_at(2);
        let mut obj_path = path::objects();
        obj_path.push(dir);
        obj_path.push(res);

        let path = match base_dir {
            Some(dir) => dir.join(name),
            None => PathBuf::from(name),
        };

        let root = path::repo_root();
        let abs_path = root.join(path.clone());

        let obj = Obj { 
            name: String::from(name),
            obj_path,
            obj_type: match obj_type {
                "tree" => ObjType::TREE,
                "blob" => ObjType::BLOB,
                _ => ObjType::DEFAULT
            },
            file_path: abs_path,
            relative_file_path: path.clone(),
            repo_file_path: path,
            hash_path: String::from(hash_path),
        };

        match obj.obj_type {
            ObjType::TREE => Box::new(Tree::new(obj)),
            ObjType::BLOB => Box::new(Blob::new(obj)),
            ObjType::DEFAULT => Box::new(Tree::new(obj))
        }
    }

    pub fn from_head() -> Self {
        Obj { 
            name: String::new(),
            obj_path: head::path(),
            obj_type: ObjType::TREE,
            file_path: PathBuf::new(),
            relative_file_path: PathBuf::new(),
            repo_file_path: PathBuf::new(),
            hash_path: String::new(),
        }
    }
}

